--[[
	Title: Fun

	Some fun things for admins
]]

function ulx.cc_slap( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A slapped #T with " .. tostring( argv[2] or 0 ) .. " damage" )
		ULib.slap( v, tonumber( argv[ 2 ] ) )
	end
end
ulx.concommand( "slap", ulx.cc_slap, "<user(s)> [<damage>] - Slaps target(s) with given damage.", ULib.ACCESS_ADMIN, "!slap", _, ulx.ID_PLAYER_HELP )

function ulx.cc_whip( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end
	
	local times = tonumber( argv[ 2 ] ) or 10
	local interval = tonumber( argv[ 3 ] ) or 0.5
	local damage = tonumber( argv[ 4 ] ) or 1

	if times > 200 then
		ULib.tsay( ply, "You cannot whip them more than 200 times!" )
		return
	end
	
	if interval <= 0 then
		ULib.tsay( ply, "Invalid interval!" )
		return
	end

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, string.format( "#A is whipping #T %i times with %i damage.", times, damage ) )
		local dtime = 0
		for i=1, times do
			timer.Simple( dtime, function()
				ULib.slap( v, damage )
			end )
			dtime = dtime + interval
		end
	end
end
ulx.concommand( "whip", ulx.cc_whip, "<user(s)> [<times>] [<interval>] [<damage>] - Slaps target(s) num times with interval between slaps.", ULib.ACCESS_ADMIN, "!whip", _, ulx.ID_PLAYER_HELP )

function ulx.cc_slay( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A slayed #T" )
		v:Kill()
	end
end
ulx.concommand( "slay", ulx.cc_slay, "<user(s)> - slays the specified users", ULib.ACCESS_ADMIN, "!slay", _, ulx.ID_PLAYER_HELP )

function ulx.cc_sslay( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A sslayed #T" )

		if v:InVehicle() then
			local vehicle = v:GetParent()
			v:ExitVehicle()
		end

		v:KillSilent()
	end
end
ulx.concommand( "sslay", ulx.cc_sslay, "<user(s)> - silently slays the specified users", ULib.ACCESS_ADMIN, "!sslay", _, ulx.ID_PLAYER_HELP )

function ulx.cc_ignite( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A ignited #T" )
		v:Ignite( 300, 100 ) -- 5 mins as max time, allow to spread
	end
end
ulx.concommand( "ignite", ulx.cc_ignite, "<user(s)> - ignites the specified users", ULib.ACCESS_ADMIN, "!ignite", _, ulx.ID_PLAYER_HELP )

local function checkFireDeath( ply )
	if ply:IsOnFire() then ply:Extinguish() end
end
hook.Add( "PlayerDeath", "ULXCheckFireDeath", checkFireDeath )

function ulx.cc_playsound( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	if not file.Exists( "../sound/" .. argv[ 1 ] ) then
		ULib.tsay( ply, "That sound doesn't exist on the server!" )
		return
	end

	ulx.logServAct( ply, "#A played sound " .. argv[ 1 ] )

	umsg.Start( "ulib_sound" )
		umsg.String( Sound( argv[ 1 ] ) )
	umsg.End()
end
ulx.concommand( "playsound", ulx.cc_playsound, "<sound> - Plays a sound, relative to 'sound' dir.", ULib.ACCESS_ADMIN )

function ulx.cc_freeze( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A froze #T" )

		if v:InVehicle() then
			local vehicle = v:GetParent()
			v:ExitVehicle()
		end

		v:Lock()
	end
end
ulx.concommand( "freeze", ulx.cc_freeze, "<user(s)> - freezes the specified users", ULib.ACCESS_ADMIN, "!freeze", _, ulx.ID_PLAYER_HELP )

function ulx.cc_unfreeze( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A unfroze #T" )
		v:UnLock()
	end
end
ulx.concommand( "unfreeze", ulx.cc_unfreeze, "<user(s)> - unfreezes the specified users", ULib.ACCESS_ADMIN, "!unfreeze", _, ulx.ID_PLAYER_HELP )

function ulx.cc_god( ply, command, argv, args )
	local targets
	if #argv < 1 then
		if not ply:IsValid() then
			Msg( "You are the console, you are already god.\n" )
			return
		end
		targets = { ply }
	else
		local err
		targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
		if not targets then
			ULib.tsay( ply, err )
			return
		end
	end

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A granted god mode to #T" )
		v:GodEnable()
	end
end
ulx.concommand( "god", ulx.cc_god, "<user(s)> - Makes player(s) invincible. If no user is specified, gods yourself.", ULib.ACCESS_ADMIN, "!god", _, ulx.ID_PLAYER_HELP )

function ulx.cc_ungod( ply, command, argv, args )
	local targets
	if #argv < 1 then
		if not ply:IsValid() then
			Msg( "Your position of god is irrevocable; if you don't like it, leave the matrix.\n" )
			return
		end
		targets = { ply }
	else
		local err
		targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
		if not targets then
			ULib.tsay( ply, err )
			return
		end
	end

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A revoked god mode from #T" )
		v:GodDisable()
	end
end
ulx.concommand( "ungod", ulx.cc_ungod, "<user(s)> - Revokes player'(s) invincibility. If no user is specified, ungods yourself.", ULib.ACCESS_ADMIN, "!ungod", _, ulx.ID_PLAYER_HELP )

function ulx.cc_noclip( ply, command, argv, args )
	local targets
	if #argv < 1 then
		if not ply:IsValid() then
			Msg( "You have no physical body to noclip.\n" )
			return
		end
		targets = { ply }
	else
		local err
		targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
		if not targets then
			ULib.tsay( ply, err )
			return
		end
	end

	for _, v in ipairs( targets ) do
		if v.jail then
			ULib.tsay( ply, v:Nick() .. " is in jail!" )
		else
			if v:GetMoveType() == MOVETYPE_WALK then
				v:SetMoveType( MOVETYPE_NOCLIP )
			elseif v:GetMoveType() == MOVETYPE_NOCLIP then
				v:SetMoveType( MOVETYPE_WALK )
			end -- Ignore if they're an observer
		end
	end
end
ulx.concommand( "noclip", ulx.cc_noclip, "<user(s)> - Toggles noclip for player(s). If no user is specified, noclips yourself.", ULib.ACCESS_ADMIN, "!noclip", _, ulx.ID_PLAYER_HELP )

function ulx.cc_hp( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end
	
	local hp = tonumber( argv[ 2 ] )
	if not hp then 
		ULib.tsay( ply, "\"" .. argv[ 2 ] .. "\" is an invalid health!" )
		return
	end

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A set #T's hp to " .. hp )
		v:SetHealth( hp )
	end
end
ulx.concommand( "hp", ulx.cc_hp, "<user(s)> - sets the hp for the specified users", ULib.ACCESS_ADMIN, "!hp", _, ulx.ID_PLAYER_HELP )

function ulx.cc_cloak( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end
	
	local amt = tonumber( argv[ 2 ] ) or 0

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A cloaked #T by amount " .. amt )
		ULib.invisible( v, true, amt )
	end
end
ulx.concommand( "cloak", ulx.cc_cloak, "<user(s)> [<amt 0-255>] - Cloaks user(s) by amount.", ULib.ACCESS_ADMIN, "!cloak", _, ulx.ID_PLAYER_HELP )

function ulx.cc_uncloak( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A uncloaked #T" )
		ULib.invisible( v, false )
	end
end
ulx.concommand( "uncloak", ulx.cc_uncloak, "<user(s)> - Uncloaks user(s).", ULib.ACCESS_ADMIN, "!uncloak", _, ulx.ID_PLAYER_HELP )

function ulx.cc_blind( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end
	
	local amt = tonumber( argv[ 2 ] ) or 255

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A blinded #T by amount " .. amt )
		umsg.Start( "ulx_blind", v )
			umsg.Bool( true )
			umsg.Short( amt )
		umsg.End()
		
		v.HadCamera = v:HasWeapon( "gmod_camera" )
		v:StripWeapon( "gmod_camera" )
	end
end
ulx.concommand( "blind", ulx.cc_blind, "<user(s)> [<amt 0-255>] - Blinds user(s) by amount.", ULib.ACCESS_ADMIN, "!blind", _, ulx.ID_PLAYER_HELP )

function ulx.cc_unblind( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A unblinded #T" )
		umsg.Start( "ulx_blind", v )
			umsg.Bool( false )
			umsg.Short( 0 )
		umsg.End()
		
		if v.HadCamera then
			v:Give( "gmod_camera" )
		end
		v.HadCamera = nil
	end
end
ulx.concommand( "unblind", ulx.cc_unblind, "<user(s)> - Unblinds user(s).", ULib.ACCESS_ADMIN, "!unblind", _, ulx.ID_PLAYER_HELP )

function ulx.cc_jail( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end

	local time = tonumber( argv[ 2 ] ) or 0

	local mdl1 = Model( "models/props_c17/fence01b.mdl" )
	local mdl2 = Model( "models/props_c17/fence02b.mdl" )
	local jail = {
		{ pos = Vector( 35, 0, 60 ), ang = Angle( 0, 0, 0 ), mdl=mdl2 },
		{ pos = Vector( -35, 0, 60 ), ang = Angle( 0, 0, 0 ), mdl=mdl2 },
		{ pos = Vector( 0, 35, 60 ), ang = Angle( 0, 90, 0 ), mdl=mdl2 },
		{ pos = Vector( 0, -35, 60 ), ang = Angle( 0, 90, 0 ), mdl=mdl2 },
		{ pos = Vector( 0, 0, 110 ), ang = Angle( 90, 0, 0 ), mdl=mdl1 },
		{ pos = Vector( 0, 0, -5 ), ang = Angle( 90, 0, 0 ), mdl=mdl1 },
	}

	for _, v in ipairs( targets ) do

		if v.jail then -- They're already jailed
			v.jail.unjail()
		end
		
		if v.ragdoll then
			ULib.tsay( ply, v:Nick() .. " is ragdolled!" )
		elseif v.ghost then
			ULib.tsay( ply, v:Nick() .. " is ghosted!" )
		else
			if time > 0 then
				ulx.logUserAct( ply, v, "#A jailed #T for " .. time .. " seconds" )
			else
				ulx.logUserAct( ply, v, "#A jailed #T" )
			end
			if v:InVehicle() then
				local vehicle = v:GetParent()
				v:ExitVehicle()
				vehicle:Remove()
			end
	
			if v:GetMoveType() == MOVETYPE_NOCLIP then -- Take them out of noclip
				v:SetMoveType( MOVETYPE_WALK )
			end
	
			local pos = v:GetPos()
	
			local walls = {}
			for _, info in ipairs( jail ) do
				local ent = ents.Create( "prop_physics" )
				ent:SetModel( info.mdl )
				ent:SetPos( pos + info.pos )
				ent:SetAngles( info.ang )
				ent:Spawn()
				ent:GetPhysicsObject():EnableMotion( false )
				table.insert( walls, ent )
			end
	
			local function unjail()
				for _, ent in ipairs( walls ) do
					ent:Remove()
				end
				v:DisallowNoclip( false )
				v:DisallowSpawning( false )
	
				v.jail = nil
			end
			if time > 0 then
				timer.Simple( time, unjail )
			end
	
			for _, ent in ipairs( walls ) do
				ent:DisallowDeleting( true )
				ent:DisallowMoving( true )
			end
			v:DisallowNoclip( true )
			v:DisallowSpawning( true )
			v.jail = { pos=pos, unjail=unjail }
		end
	end
end
ulx.concommand( "jail", ulx.cc_jail, "<user(s)> [<time>] - Jails user(s) for specified seconds (forever if not specified).", ULib.ACCESS_ADMIN, "!jail", _, ulx.ID_PLAYER_HELP )

function ulx.cc_unjail( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end

	for _, v in ipairs( targets ) do
		if v.jail then
			ulx.logUserAct( ply, v, "#A unjailed #T" )
			v.jail.unjail()
			v.jail = nil
		end
	end
end
ulx.concommand( "unjail", ulx.cc_unjail, "<user(s)> - Unjails user(s).", ULib.ACCESS_ADMIN, "!unjail", _, ulx.ID_PLAYER_HELP )

function ulx.cc_bring( ply, command, argv, args )
	if not ply:IsValid() then
		ULib.tsay( ply, "You can't use this from a dedicated console!" )
		return
	end

	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local target, err = ULib.getUser( argv[ 1 ], _, ply )
	if not target then
		ULib.tsay( ply, err )
		return
	end
	
	if target == ply then
		ULib.tsay( ply, "You can't bring yourself!" )
		return
	end
	
	if ply.jail then
		ULib.tsay( ply, "You are in jail!" )
		return
	end

	if target.jail then
		ULib.tsay( ply, "That user is in jail!" )
		return
	end
	
	if target.ragdoll then
		ULib.tsay( ply, "That user is ragdolled!" )
		return
	end
	
	local ang = ply:GetAimVector()
	ang.z = 0
	local newpos = ply:GetPos() + ang*50
	if not util.IsInWorld( newpos ) or not util.IsInWorld( newpos + ang*17 ) then -- Let's try behind
		newpos = ply:GetPos() - ang*50
		if not util.IsInWorld( newpos ) or not util.IsInWorld( newpos + ang*17 ) then
			ULib.tsay( ply, "Can't find a place to put them!" )
			return
		end
	end
	
	if target:InVehicle() then
		target:ExitVehicle()
	end

	local newang = ply:EyeAngles()
	newang.pitch = 0
	newang.roll = 0	
	newang.yaw = newang.yaw + 180

	target:SetPos( newpos )
	target:SetEyeAngles( newang )
	
	ulx.logUserAct( ply, target, "#A brought #T to him/her" )
end
ulx.concommand( "bring", ulx.cc_bring, "<user> - Brings a user to you", ULib.ACCESS_ADMIN, "!bring", _, ulx.ID_PLAYER_HELP )

function ulx.cc_goto( ply, command, argv, args )
	if not ply:IsValid() then
		ULib.tsay( ply, "You can't use this from a dedicated console!" )
		return
	end

	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local target, err = ULib.getUser( argv[ 1 ], true, ply ) -- Ignore immunity
	if not target then
		ULib.tsay( ply, err )
		return
	end
	
	if target == ply then
		ULib.tsay( ply, "You can't goto yourself!" )
		return
	end
	
	if ply.jail then
		ULib.tsay( ply, "You are in jail!" )
		return
	end
	
	if ply.ragdoll then
		ULib.tsay( ply, "You are ragdolled!" )
		return
	end
	
	local ang = target:GetAimVector()
	ang.z = 0
	local newpos = target:GetPos() + ang*50
	if not util.IsInWorld( newpos ) or not util.IsInWorld( newpos + ang*17 ) then -- Let's try behind
		newpos = target:GetPos() - ang*50
		if not util.IsInWorld( newpos ) or not util.IsInWorld( newpos + ang*17 ) then
			ULib.tsay( ply, "Can't find a place to put you!" )
			return
		end
	end
	
	if ply:InVehicle() then
		ply:ExitVehicle()
	end

	local newang = target:EyeAngles()
	newang.pitch = 0
	newang.roll = 0
	newang.yaw = newang.yaw + 180

	ply:SetPos( newpos )
	ply:SetEyeAngles( newang )
	ulx.logUserAct( ply, target, "#A teleported to #T" )
end
ulx.concommand( "goto", ulx.cc_goto, "<user> - Goto a user", ULib.ACCESS_ADMIN, "!goto", _, ulx.ID_PLAYER_HELP )

function ulx.cc_teleport( ply, command, argv, args )
	if not ply:IsValid() then
		ULib.tsay( ply, "You can't use this from a dedicated console!" )
		return
	end
	
	if ply:InVehicle() then
		ULib.tsay( ply, "Leave the vehicle first!" )
		return
	end
	
	if ply.jail then
		ULib.tsay( ply, "You are in jail!" )
		return
	end

	local tr = ply:GetEyeTrace()
	local pos = tr.HitPos - ply:GetAimVector()*72 -- Make sure they don't end up in world

	ply:SetPos( pos )
end
ulx.concommand( "teleport", ulx.cc_teleport, " - Teleports you to where you're looking.", ULib.ACCESS_ADMIN, "!tp", true, ulx.ID_HELP )

local function doGhost( ply )
	if not ply.ghost then return end
	local dTime = FrameTime()
	local speed = 500
	local upspeed = 200

	if ply:KeyDown( IN_FORWARD ) then
		ply:SetPos( ply:GetPos() + ply:GetForward() * (speed*dTime) )
	elseif ply:KeyDown( IN_BACK ) then
		ply:SetPos( ply:GetPos() + ply:GetForward() * (speed*dTime*-1) )
	end

	if ply:KeyDown( IN_MOVERIGHT ) then
		ply:SetPos( ply:GetPos() + ply:GetRight() * (speed*dTime) )
	elseif ply:KeyDown( IN_MOVELEFT ) then
		ply:SetPos( ply:GetPos() + ply:GetRight() * (speed*dTime*-1) )
	end

	if ply:KeyDown( IN_JUMP ) then
		ply:SetPos( ply:GetPos() + ply:GetUp() * (upspeed*dTime) )
	elseif ply:KeyDown( IN_DUCK ) then
		ply:SetPos( ply:GetPos() + ply:GetUp() * (upspeed*dTime*-1) )
	end
end

function ulx.cc_ghost( ply, command, argv, args )
	local targets
	if #argv < 1 then
		if not ply:IsValid() then
			Msg( "You have no physical body to ghost.\n" )
			return
		end
		targets = { ply }
	else
		local err
		targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
		if not targets then
			ULib.tsay( ply, err )
			return
		end
	end
	
	for _, v in ipairs( targets ) do
		if v:InVehicle() then
			v:ExitVehicle()
		end
		
		if v.jail then
			ULib.tsay( ply, v:Nick() .. " is in jail!" )
		elseif v.ragdoll then
			ULib.tsay( ply, v:Nick() .. " is ragdolled!" )
		else
			v:SetMoveType( MOVETYPE_NONE )
			v:SetNotSolid( true )
			v:GodEnable()
			ULib.invisible( v, true )
			v.ghost = true

			-- This is a self-removing think for ghosts. When there's no more ghosts, it dies.		
			local function ghostThink()
				local remove = true
				local players = player.GetAll()
				for _, player in ipairs( players ) do
					if player.ghost then
						doGhost( player )
						remove = false
					end
				end
				if remove then
					hook.Remove( "Think", "ULXGhostThink" )
				end
			end
			hook.Add( "Think", "ULXGhostThink", ghostThink )
			
			ulx.logUserAct( ply, v, "#A ghosted #T", true )
		end
	end
end
ulx.concommand( "ghost", ulx.cc_ghost, " - Ghosts you.", ULib.ACCESS_ADMIN, "!ghost", true, ulx.ID_HELP )

function ulx.cc_unghost( ply, command, argv, args )
	local targets
	if #argv < 1 then
		if not ply:IsValid() then
			Msg( "You have no physical body to unghost.\n" )
			return
		end
		targets = { ply }
	else
		local err
		targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
		if not targets then
			ULib.tsay( ply, err )
			return
		end
	end
	
	for _, v in ipairs( targets ) do
		if v:InVehicle() then
			v:ExitVehicle()
		end

		if v.ragdoll then
			ULib.tsay( ply, v:Nick() .. " is ragdolled!" )
		else
			v:SetMoveType( MOVETYPE_WALK )
			v:SetNotSolid( false )
			v:GodDisable()
			ULib.invisible( v, false )
			v.ghost = nil

			ulx.logUserAct( ply, v, "#A unghosted #T", true )
		end
	end
end
ulx.concommand( "unghost", ulx.cc_unghost, " - Unghosts you.", ULib.ACCESS_ADMIN, "!unghost", true, ulx.ID_HELP )

local function jailSpawnCheck( ply )
	if ply.jail then
		ply:SetPos( ply.jail.pos )
	end
end
hook.Add( "PlayerSpawn", "ULXJailSpawnCheck", jailSpawnCheck )

local function jailDisconnectedCheck( ply )
	if ply.jail then
		ply.jail.unjail()
	end
end
hook.Add( "PlayerDisconnected", "ULXJailDisconnectedCheck", jailDisconnectedCheck )

function ulx.cc_ragdoll( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end

	for _, v in ipairs( targets ) do
		if v.ragdoll then -- They're already ragdolled, what do you want?
		elseif v.jail then
			ULib.tsay( ply, v:Nick() .. " is in jail!" )
		else
			ulx.logUserAct( ply, v, "#A ragdolled #T" )
			
			if v:InVehicle() then
				local vehicle = v:GetParent()
				v:ExitVehicle()
			end
	
			local ragdoll = ents.Create( "prop_ragdoll" )
			ragdoll:SetPos( v:GetPos() )
			ragdoll:SetModel( v:GetModel() )
			ragdoll:Spawn()
			ragdoll:Activate()
	
			v:Spectate( OBS_MODE_CHASE )
			v:SpectateEntity( ragdoll )
			v:StripWeapons() -- Otherwise they can still use the weapons.
	
			ragdoll:DisallowDeleting( true )
			v:DisallowSpawning( true )

			v.ragdoll = ragdoll
		end
	end
end
ulx.concommand( "ragdoll", ulx.cc_ragdoll, "<user(s)> - ragdolls the specified users", ULib.ACCESS_ADMIN, "!ragdoll", _, ulx.ID_PLAYER_HELP )

function ulx.cc_unragdoll( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end

	for _, v in ipairs( targets ) do
		if v.ragdoll then -- Only if they're ragdolled...
			ulx.logUserAct( ply, v, "#A unragdolled #T" )
			local pos = v.ragdoll:GetPos()
			pos.z = pos.z + 10 -- So they don't end up in the ground
			v:DisallowSpawning( false )
			v.ragdoll:Remove()
			v.ragdoll = nil -- Gotta do this before spawn or our hook catches it
			v:Spawn() -- Take them out of observer mode, and give them weapons back.
			v:SetPos( pos )
		end
	end
end
ulx.concommand( "unragdoll", ulx.cc_unragdoll, "<user(s)> - unragdolls the specified users", ULib.ACCESS_ADMIN, "!unragdoll", _, ulx.ID_PLAYER_HELP )

local function ragdollSpawnCheck( ply )
	if ply.ragdoll then
		timer.Simple( 0.01, function() -- Doesn't like us using it instantly
			ply:Spectate( OBS_MODE_CHASE )
			ply:SpectateEntity( ply.ragdoll )
			ply:StripWeapons() -- Otherwise they can still use the weapons.
		end )
	end
end
hook.Add( "PlayerSpawn", "ULXRagdollSpawnCheck", ragdollSpawnCheck )

local function ragdollDisconnectedCheck( ply )
	if ply.ragdoll then
		ply.ragdoll:Remove()
	end
end
hook.Add( "PlayerDisconnected", "ULXRagdollDisconnectedCheck", ragdollDisconnectedCheck )

local function voteDone( t )
	local results = t.results
	local winner
	local winnernum = 0
	for id, numvotes in pairs( results ) do
		if numvotes > winnernum then
			winner = id
			winnernum = numvotes
		end
	end

	local str
	if not winner then
		str = "Vote results: No option won because no one voted!"
	else
		str = "Vote results: Option '" .. t.options[ winner ] .. "' won. (" .. winnernum .. "/" .. t.voters .. ")"
	end
	ULib.tsay( _, str )
	ulx.logString( str )
	Msg( str .. "\n" )
end

function ulx.cc_vote( ply, command, argv, args )
	if #argv < 3 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local title = argv[ 1 ]
	table.remove( argv, 1 )
	ulx.doVote( title, argv, voteDone )
	ulx.logServAct( ply, "#A started a vote." )
end
ulx.concommand( "vote", ulx.cc_vote, "<title> <option1> <option2> [<option3>] .. - Starts a public vote", ULib.ACCESS_ALL, "!vote", _, ulx.ID_HELP )

local gimpSays = {} -- Holds gimp says
local ID_GIMP = 1
local ID_MUTE = 2

function ulx.cc_gimp( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A gimped #T" )
		v.gimp = ID_GIMP
	end
end
ulx.concommand( "gimp", ulx.cc_gimp, "<user> - Gimps a player (can't talk normally)", ULib.ACCESS_ADMIN, "!gimp", _, ulx.ID_PLAYER_HELP )

function ulx.cc_mute( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A muted #T" )
		v.gimp = ID_MUTE
	end
end
ulx.concommand( "mute", ulx.cc_mute, "<user> - Mutes a player (can't talk)", ULib.ACCESS_ADMIN, "!mute", _, ulx.ID_PLAYER_HELP )

function ulx.cc_ungimp( ply, command, argv, args )
	if #argv < 1 then
		ULib.tsay( ply, ulx.LOW_ARGS )
		return
	end

	local targets, err = ULib.getUsers( argv[ 1 ], _, true, ply ) -- Enable keywords
	if not targets then
		ULib.tsay( ply, err )
		return
	end

	for _, v in ipairs( targets ) do
		ulx.logUserAct( ply, v, "#A ungimped/unmuted #T" )
		v.gimp = nil
	end
end
ulx.concommand( "ungimp", ulx.cc_ungimp, "<user> - Ungimps a player", ULib.ACCESS_ADMIN, "!ungimp", _, ulx.ID_PLAYER_HELP )
ulx.concommand( "unmute", ulx.cc_ungimp, "<user> - Unmutes a player", ULib.ACCESS_ADMIN, "!unmute", _, ulx.ID_PLAYER_HELP )

function ulx.cc_addGimpSay( ply, command, argv, args )
	local say = ULib.stripQuotes( args )
	table.insert( gimpSays, say )
end
ulx.concommand( "addGimpSay", ulx.cc_addGimpSay, "", ULib.ACCESS_NONE ) -- Can only add from configs


---------------------
--      Hooks      --
---------------------


local function gimpCheck( ply, strText )
	if ply.gimp == ID_MUTE then return "" end
	if ply.gimp == ID_GIMP then
		if #gimpSays < 1 then return nil end
		return gimpSays[ math.random( #gimpSays ) ]
	end
end
hook.Add( "PlayerSay", "ULXGimpCheck", gimpCheck )

local function meCheck( ply, strText, bPublic )
	if strText:sub( 1, 4 ) == "/me " then
		strText = string.format( "*** %s %s", ply:Nick(), strText:sub( 5 ) )
		if bPublic then
			ULib.tsay( _, strText )
		else
			strText = "(TEAM) " .. strText
			local teamid = ply:Team()
			local players = team.GetPlayers( teamid )
			for _, ply2 in ipairs( players ) do
				ULib.tsay( ply2, strText )
			end
		end
		
		if isDedicatedServer() then
			Msg( strText .. "\n" ) -- Log to console
		end
		if util.tobool( GetConVarNumber( "ulx_logChat" ) ) then
			ulx.logString( strText )
		end
		
		return ""
	end
end
hook.Add( "PlayerSay", "ULXMeCheck", meCheck )